import logging
from datetime import datetime, timedelta
from botocore.exceptions import ClientError
from org_retention.lambda_handler import setup_lambda_handler
from org_retention.utils.config import Config
from pscr_common.aws import get_boto3_resource

DEFAULT_RULE_OFFSET = (
    0  # number of days to offset the effective expiration date of the rules
)
DEFAULT_RULES_COUNT = 7  # number of rules to create up until the starting date
DATE_FORMAT = "%Y-%m-%d"

_logger = logging.getLogger(__name__)


@setup_lambda_handler(Config())
def handler(event, _context):
    now = datetime.now()
    _logger.info(
        "Initializing S3 retention rules lambda with event: %s and local time: %s",
        event,
        now,
    )

    # Extract configuration from invocation event
    bucket = event.get("s3_bucket")  # Mandatory argument
    prefix = event.get("s3_prefix", "")
    offset = int(event.get("retention_rule_offset", DEFAULT_RULE_OFFSET)) + 1
    count = int(event.get("retention_rules_count", DEFAULT_RULES_COUNT))
    rules = event.get("static_rules", [])

    # Update the bucket lifecycle policy
    new_rules = _static_rules(rules) + _dynamic_rules(now, prefix, offset, count)
    lifecycle_config = _load_lifecycle_config(bucket)
    lifecycle_config.put(LifecycleConfiguration={"Rules": new_rules})
    lifecycle_config.reload()  # Reload in order to print the changed rules

    _logger.info(
        "Lifecycle rules after updating retentions: %s", lifecycle_config.rules
    )


def _static_rules(rules):
    return [{**rule, "ID": f"static-rule-{i}"} for i, rule in enumerate(rules)]


def _dynamic_rules(start_date, prefix, offset, rules_count):
    dates = [(start_date - timedelta(days=i)) for i in range(rules_count)]
    return [
        {
            "ID": f"org-retention-{date.strftime(DATE_FORMAT)}",
            "Expiration": {
                "Date": (date + timedelta(days=offset)).strftime(DATE_FORMAT)
            },
            "Filter": {"Prefix": f"{prefix}expire={date.strftime(DATE_FORMAT)}/"},
            "Status": "Enabled",
        }
        for date in dates
    ]


def _load_lifecycle_config(bucket):
    lifecycle_config = get_boto3_resource("s3").BucketLifecycleConfiguration(bucket)
    try:
        _logger.info(
            "Lifecycle rules before updating retentions: %s", lifecycle_config.rules
        )
    except ClientError as error:
        if "NoSuchLifecycleConfiguration" in str(error):
            _logger.info("Bucket does not have LifecycleConfiguration set.")
        else:
            raise error
    return lifecycle_config
