import asyncio
import hashlib
import logging
from datetime import datetime
from typing import Optional, Dict, List

import rapidjson as json
from botocore.exceptions import ClientError

from org_retention.api.models import RetentionConfig
from org_retention.utils.config import Config
from org_retention.utils.tags import Tags
from pscr_common.aws import get_boto3_async_client as client

_logger = logging.getLogger(__name__)


class S3Client(object):
    def __init__(self, config: Config):
        self.config = config

        self.retention_bucket = config.get("persist.s3.retention", "bucket")
        self.retention_key = config.get("persist.s3.retention", "key")
        self.retention_prefix = config.get("persist.s3.retention", "prefix")
        self.retention_suffix = config.get("persist.s3.retention", "suffix")

        self.retention_config_bucket = config.get(
            "persist.s3.retention_config", "bucket"
        )
        self.retention_config_latest_key = config.get(
            "persist.s3.retention_config", "latest_key"
        )
        self.retention_config_tstamp_key = config.get(
            "persist.s3.retention_config", "tstamp_key"
        )

    async def get_retention_days(
        self, org_key: str, retention_class: str
    ) -> Optional[int]:
        key = self.retention_key.format(org_key, retention_class)
        config = await self._get_json_from_key(key)
        return config["days"] if config else None

    async def set_retention_days(
        self, org_key: str, retention_class: str, days: int
    ) -> None:
        async with client("s3") as s3:
            key = self.retention_key.format(org_key, retention_class)
            body = json.dumps({"org": org_key, "class": retention_class, "days": days})
            await s3.put_object(Bucket=self.retention_bucket, Key=key, Body=body)
            _logger.info(
                "Retention for org %s %s class set to %s days",
                org_key,
                retention_class,
                days,
            )

    async def clear_retention_days(self, org_key: str, retention_class: str) -> None:
        async with client("s3") as s3:
            key = self.retention_key.format(org_key, retention_class)
            await s3.delete_object(Bucket=self.retention_bucket, Key=key)
            _logger.info(
                "Retention for org %s %s class cleared", org_key, retention_class
            )

    async def get_all_org_retentions(self) -> List[RetentionConfig]:
        retentions = []

        async for key in self._get_matching_s3_keys(
            self.retention_bucket, self.retention_prefix, self.retention_suffix
        ):
            try:
                config = await self._get_json_from_key(key)
                retentions.append(
                    RetentionConfig(
                        org=config["org"],
                        retention_class=config["class"],
                        retention_days=config["days"],
                    )
                )
            except:
                _logger.warning("Invalid retention config at: %s", key)

        return retentions

    async def set_retention_config_file_tags(self, tags: Tags):
        """Sets the object tags for the retention config file, creates an empty file if non-existent"""

        async with client("s3") as s3:
            try:
                await s3.put_object_tagging(
                    Bucket=self.retention_config_bucket,
                    Key=self.retention_config_latest_key,
                    Tagging=tags.to_tag_set(),
                )
            except s3.exceptions.NoSuchKey:
                _logger.warning(
                    "Custom retention org list not found! Creating an empty object.. key: %s bucket: %s",
                    self.retention_config_latest_key,
                    self.retention_config_bucket,
                )
                await s3.put_object(
                    Bucket=self.retention_config_bucket,
                    Key=self.retention_config_latest_key,
                    Tagging=tags.to_str(),
                    Body="[]",
                )

    async def get_retention_config_file_tags(self) -> Tags:
        async with client("s3") as s3:
            try:
                resp = await s3.get_object_tagging(
                    Bucket=self.retention_config_bucket,
                    Key=self.retention_config_latest_key,
                )

                return Tags.from_tag_set(resp.get("TagSet", []))
            except s3.exceptions.NoSuchKey:
                return Tags()

    async def get_retention_config_file_md5(self):
        async with client("s3") as s3:
            try:
                resp = await s3.head_object(
                    Bucket=self.retention_config_bucket,
                    Key=self.retention_config_latest_key,
                )

                return resp.get("Metadata", {}).get("md5sum")
            except s3.exceptions.NoSuchKey:
                return None

    async def write_retention_config_file(
        self, retention_config: Dict[str, Dict[str, int]]
    ):
        """
        Writes the given retention config to a latest and timestamped object to preconfigured keys in S3
        Checks MD5 sum of previous file to avoid overwriting with the same content
        """
        async with client("s3") as s3:
            body = json.dumps(retention_config).encode("utf-8")
            md5 = hashlib.md5(body, usedforsecurity=False).hexdigest()
            if md5 != await self.get_retention_config_file_md5():
                await asyncio.gather(
                    s3.put_object(
                        Bucket=self.retention_config_bucket,
                        Body=body,
                        Key=self.retention_config_tstamp_key.format(
                            datetime.now().isoformat()
                        ),
                    ),
                    s3.put_object(
                        Bucket=self.retention_config_bucket,
                        Body=body,
                        Key=self.retention_config_latest_key,
                        Metadata={"md5sum": md5},
                    ),
                )
            else:
                _logger.info(
                    "Skipped S3 overwrite of retention config since MD5 did not change. MD5: %s",
                    md5,
                )

    async def _get_json_from_key(self, key: str) -> Optional[Dict]:
        async with client("s3") as s3:
            try:
                obj = await s3.get_object(Bucket=self.retention_bucket, Key=key)
                return json.loads(await obj["Body"].read())
            except ClientError as ex:
                if ex.response["Error"]["Code"] == "NoSuchKey":
                    return None
                raise Exception(f"Unable to get retention config: {str(ex)}")

    async def _get_matching_s3_keys(self, bucket, prefix=None, suffix=None):
        async with client("s3") as s3:
            async for obj in self._get_matching_s3_objects(s3, bucket, prefix, suffix):
                yield obj["Key"]

    @staticmethod
    async def _get_matching_s3_objects(s3, bucket, prefix=None, suffix=None):
        kwargs = {"Bucket": bucket}

        if prefix:
            kwargs["Prefix"] = prefix

        more_keys = True

        while more_keys:
            resp = await s3.list_objects_v2(**kwargs)

            contents = resp.get("Contents", [])

            for obj in contents:
                key = obj["Key"]
                if not suffix or key.endswith(suffix):
                    yield obj

            if resp["IsTruncated"]:
                more_keys = True
                kwargs["ContinuationToken"] = resp["NextContinuationToken"]
            else:
                more_keys = False
