import json
import logging
from datetime import datetime, timedelta
from os import environ as env

import botocore
from pscr_common.log import clear_logging_handlers, setup_logging

from org_retention.utils.config import Config
from pscr_common.aws import get_boto3_client

# number of days to offset the effective expiration date of the rules
DEFAULT_RULE_OFFSET = 0
# number of rules to create up until the starting date
DEFAULT_RULES_COUNT = 7
DATE_FORMAT = "%Y-%m-%d"

_logger = logging.getLogger(__name__)


class S3LifeCycleRules:
    s3_client = get_boto3_client("s3")
    now = datetime.now()

    # S3 Solr Index Bucket Config
    s3_index_bucket = env.get("S3_INDEX_BUCKET")
    s3_index_prefix = env.get("S3_INDEX_PREFIX", "")
    s3_index_offset = (
        int(env.get("S3_INDEX_RETENTION_RULE_OFFSET", DEFAULT_RULE_OFFSET)) + 1
    )
    s3_index_count = int(env.get("S3_INDEX_RETENTION_RULES_COUNT", DEFAULT_RULES_COUNT))
    s3_index_rules = env.get("S3_INDEX_STATIC_RULES", "[]")

    # S3 Solr Document Bucket Config
    s3_document_bucket = env.get("S3_DOCUMENT_BUCKET")
    s3_document_prefix = env.get("S3_DOCUMENT_PREFIX", "")
    s3_document_offset = (
        int(env.get("S3_DOCUMENT_RETENTION_RULE_OFFSET", DEFAULT_RULE_OFFSET)) + 1
    )
    s3_document_count = int(
        env.get("S3_DOCUMENT_RETENTION_RULES_COUNT", DEFAULT_RULES_COUNT)
    )
    s3_document_rules = env.get("S3_DOCUMENT_STATIC_RULES", "[]")
    s3_document_partition = int(env.get("S3_DOCUMENT_PARTITION", 16))

    @staticmethod
    def update_s3_lifecycle_rules(
        s3_bucket, rules, now, prefix, offset, count, partition=None
    ):
        lc_rules = S3LifeCycleRules.get_s3_lifecycle_rules(s3_bucket)
        _logger.info(
            f"S3 Bucket {s3_bucket} Lifecycle rules before updating retentions: {lc_rules}"
        )
        updated_rules = S3LifeCycleRules.build_new_lc_rules(
            rules, now, prefix, offset, count, partition
        )
        S3LifeCycleRules.s3_client.put_bucket_lifecycle_configuration(
            Bucket=s3_bucket,
            LifecycleConfiguration={"Rules": updated_rules},
            ChecksumAlgorithm="SHA256",
        )
        lc_rules = S3LifeCycleRules.get_s3_lifecycle_rules(s3_bucket)

        _logger.info(
            f"S3 Bucket {s3_bucket} Lifecycle rules after updating retentions: {lc_rules}",
        )

    @staticmethod
    def get_s3_lifecycle_rules(s3_bucket):
        try:
            return S3LifeCycleRules.s3_client.get_bucket_lifecycle_configuration(
                Bucket=s3_bucket
            )["Rules"]
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "NoSuchLifecycleConfiguration":
                _logger.info(
                    f"Bucket {s3_bucket} does not have LifecycleConfiguration set."
                )
            else:
                raise error
        except Exception as e:
            print(e)

        return []

    @staticmethod
    def build_new_lc_rules(rules, now, prefix, offset, count, partition=None):
        return (
            S3LifeCycleRules.get_static_rules(json.loads(rules))
            + S3LifeCycleRules.get_dynamic_rules(now, prefix, offset, count)
            + (
                S3LifeCycleRules.get_dynamic_rules(
                    now, prefix, offset, count, partition
                )
                if partition
                else []
            )
        )

    @staticmethod
    def get_static_rules(rules):
        return [{**rule, "ID": f"static-rule-{i}"} for i, rule in enumerate(rules)]

    @staticmethod
    def get_dynamic_rules(start_date, prefix, offset, rules_count, partition=None):
        dates = [(start_date - timedelta(days=i)) for i in range(rules_count)]

        s3_lc_rules = []

        if partition:
            for p in range(partition):
                partition_key = "{0:x}".format(p)
                for date in dates:
                    rule = {
                        "ID": f"org-retention-{partition_key}-{date.strftime(DATE_FORMAT)}",
                        "Expiration": {
                            "Date": (date + timedelta(days=offset)).strftime(
                                DATE_FORMAT
                            )
                        },
                        "Filter": {
                            "Prefix": f"{prefix}partition={partition_key}/expire={date.strftime(DATE_FORMAT)}/"
                        },
                        "Status": "Enabled",
                    }
                    s3_lc_rules.append(rule)
        else:
            s3_lc_rules = [
                {
                    "ID": f"org-retention-{date.strftime(DATE_FORMAT)}",
                    "Expiration": {
                        "Date": (date + timedelta(days=offset)).strftime(DATE_FORMAT)
                    },
                    "Filter": {
                        "Prefix": f"{prefix}expire={date.strftime(DATE_FORMAT)}/"
                    },
                    "Status": "Enabled",
                }
                for date in dates
            ]

        return s3_lc_rules


def main():
    config = Config()

    clear_logging_handlers()
    setup_logging(
        level=config.get("log", "level"),
        svc_version=env.get("CONTAINERVERSION", "UNKNOWN"),
        env_id=env.get("ENV_ID", "UNKNOWN"),
        svc_id="TODO",
    )

    # Update S3 Index Life Cycle Rules
    S3LifeCycleRules.update_s3_lifecycle_rules(
        S3LifeCycleRules.s3_index_bucket,
        S3LifeCycleRules.s3_index_rules,
        S3LifeCycleRules.now,
        S3LifeCycleRules.s3_index_prefix,
        S3LifeCycleRules.s3_index_offset,
        S3LifeCycleRules.s3_index_count,
    )

    # Update S3 Documents Life Cycle Rules
    S3LifeCycleRules.update_s3_lifecycle_rules(
        S3LifeCycleRules.s3_document_bucket,
        S3LifeCycleRules.s3_document_rules,
        S3LifeCycleRules.now,
        S3LifeCycleRules.s3_document_prefix,
        S3LifeCycleRules.s3_document_offset,
        S3LifeCycleRules.s3_document_count,
        S3LifeCycleRules.s3_document_partition,
    )


if __name__ == "__main__":
    main()
