import asyncio
import logging
from typing import Callable

from aiohttp import web
from aio_pyformance.registry import MetricsRegistry
from aiohttp.web_exceptions import (
    HTTPException,
    HTTPInternalServerError,
    HTTPBadRequest,
    HTTPNotFound,
    HTTPServiceUnavailable,
)

from pscr_common.api.models import ValidationError, TranslationTypeError

_logger = logging.getLogger(__name__)


def get_route_metric_label(route) -> str:
    if route.resource is None:
        return "unknown_route"
    parts = [p.strip("/") for p in route.resource.canonical.split("/") if len(p) > 0]
    parts = [
        p.strip("{}").replace("_", "").upper() if p.startswith("{") else p
        for p in parts
    ]  # Upper variable names
    return "_".join(parts) + "." + route.method.lower()


def error_middleware_factory(metrics_registry: MetricsRegistry):
    """
    This middleware factory creates middlewares that will log metrics for any non-200 response
    :param metrics_registry:
    :return:
    """

    @web.middleware
    async def error_middleware(request: web.Request, handler: Callable) -> web.Response:
        request_metric_prefix = (
            f"requests.{get_route_metric_label(request.match_info.route)}"
        )
        response_status = 200

        try:
            with metrics_registry.timer(request_metric_prefix + ".time").time():
                response = await handler(request)
                response_status = response.status
                return response
        except ValidationError as ex:
            _logger.info("Error validating request: %s", str(ex))
            response_status = 400
            raise HTTPBadRequest(reason=str(ex))
        except HTTPBadRequest as ex:
            _logger.info("Error processing request: %s", str(ex))
            response_status = ex.status
            raise
        except HTTPNotFound as ex:
            _logger.info("Error processing request: %s", str(ex))
            response_status = ex.status
            raise
        except HTTPException as ex:
            _logger.exception("Error during request")
            response_status = ex.status
            raise
        except ValueError as ex:
            _logger.info(f"Value error during request: {str(ex)}")
            response_status = 400
            raise HTTPBadRequest(reason=str(ex))
        except TranslationTypeError as ex:
            _logger.info(f"Type error parsing request: {str(ex)}")
            response_status = 400
            raise HTTPBadRequest(reason=str(ex))
        except asyncio.CancelledError:
            _logger.info("Request cancelled due to shutdown: %s", request)
            response_status = 503
            raise HTTPServiceUnavailable()
        except Exception:
            _logger.exception(f"Unknown error during request {str(request)}")
            response_status = 500
            raise HTTPInternalServerError()
        finally:
            if response_status >= 400:
                await metrics_registry.meter(
                    request_metric_prefix + f".errors.{response_status}"
                ).mark()

    return error_middleware
