import logging

import rapidjson
from aiohttp import web
from datetime import datetime
import json
import os

from org_retention.api.models import Retention
from org_retention.engine.retention_engine import RetentionEngine
from pscr_common.web.auth import AuthRouteTableDef, Action
from pscr_common.web.auditlog import call_audit_log

from org_retention.utils import json_dumps
from org_retention.web.auth import Resource

_logger = logging.getLogger(__name__)


def make_response_json(data=None) -> web.Response:
    if data is None:
        return web.Response(status=204)
    else:
        return web.json_response(data, dumps=json_dumps)


def build_route_table(engine: RetentionEngine) -> AuthRouteTableDef:
    routes = AuthRouteTableDef()

    @routes.get("/healthcheck")
    async def healthcheck(request: web.Request):
        """
        ---
        summary: Healthcheck
        description: Successful response indicates service reachability.
        produces:
        - application/octet-stream
        responses:
            "204":
                description: service is available
                content:
                  "*:*": {}
        """
        request.do_not_log = True

        return make_response_json()

    @routes.get("/v1/allowed_days", Resource.internal, [Action.execute])
    async def get_allowed_days(_request: web.Request):
        return make_response_json(await engine.get_retention_class_allowed_days())

    @routes.get("/v1/orgs/{org_key}", Resource.provisioning, [Action.read])
    async def get_org_config(request: web.Request):
        """
        ---
        summary: Get org config
        description: Retrieve complete configuration for org
        produces:
        - application/json
        responses:
            "200":
                description: Org config
                content:
                  "application/json":
                    schema:
                      $ref: '#/definitions/OrgConfig'
        """
        org_key = request.match_info["org_key"]

        return make_response_json(await engine.get_org_config(org_key))

    @routes.get("/v1/orgs/{org_key}/classes", Resource.retention, [Action.read])
    async def get_all_retentions(request: web.Request):
        """
        ---
        summary: Get all class retentions
        description: Retrieve retention for all classes.
        produces:
        - application/json
        responses:
            "200":
                description: Events retention
                content:
                  "application/json":
                    schema:
                      $ref: '#/definitions/Retention'
            "404":
                description: Unknown retention class
                content:
                  "*:*": {}
        """
        org_key = request.match_info["org_key"]

        return make_response_json(await engine.get_all_retentions_for_org(org_key))

    @routes.get(
        "/v1/orgs/{org_key}/classes/{retention_class}",
        Resource.retention,
        [Action.read],
    )
    async def get_retention(request: web.Request):
        """
        ---
        summary: Get class retention
        description: Retrieve retention for given class.
        produces:
        - application/json
        responses:
            "200":
                description: Events retention
                content:
                  "application/json":
                    schema:
                      $ref: '#/definitions/Retention'
            "404":
                description: Unknown retention class
                content:
                  "*:*": {}
        """
        retention_class = request.match_info["retention_class"]
        org_key = request.match_info["org_key"]

        return make_response_json(await engine.get_retention(org_key, retention_class))

    @routes.put(
        "/v1/orgs/{org_key}/classes/{retention_class}",
        Resource.provisioning,
        [Action.update],
    )
    async def set_retention(request: web.Request):
        """
        ---
        summary: Set class retention
        description: Update org retention for class in days
        produces:
        - application/json
        responses:
            "200":
                description: Events retention
                content:
                  "application/json":
                    schema:
                      $ref: '#/definitions/Retention'
            "404":
                description: Unknown retention class
                content:
                  "*:*": {}
            "422":
                description: Non-allowed retention days input
                content:
                  "*:*": {}
        """
        retention_class = request.match_info["retention_class"]
        org_key = request.match_info["org_key"]

        retention_days = request.query.get("retention_days")
        if retention_days is not None:
            retention_days = int(retention_days)
        else:
            params = await request.json(loads=rapidjson.loads)
            retention_days = Retention.from_dict(params).retention_days

        if os.environ.get("ENABLE_AUDITLOG").lower() == "true":
            call_audit_log(
                request,
                f"Received request to update retention class { retention_class } to { retention_days } days",
            )

        return make_response_json(
            await engine.set_retention(org_key, retention_class, retention_days)
        )

    @routes.delete(
        "/v1/orgs/{org_key}/classes/{retention_class}",
        Resource.provisioning,
        [Action.update],
    )
    async def clear_retention(request: web.Request):
        """
        ---
        summary: Clear class retention
        description: Clears org custom retention settings
        responses:
            "204":
                description: Events retention
                content:
                  "*:*": {}
            "404":
                description: Unknown retention class
                content:
                  "*:*": {}
        """
        retention_class = request.match_info["retention_class"]
        org_key = request.match_info["org_key"]

        if os.environ.get("ENABLE_AUDITLOG").lower() == "true":
            call_audit_log(
                request,
                f"Clearing data retention settings for retention class { retention_class }",
            )

        await engine.set_default_retention(org_key, retention_class)
        return make_response_json()

    return routes
